@@ -44,6 +44,7 @@
 #include "htmltext.h"
 #include "htmlrule.h"
 #include "htmltype.h"
+#include "htmlsettings.h"
 
 #include "gtkhtmldebug.h"
 
@@ -64,7 +65,9 @@ destroy (HTMLObject *self)
 	self->next = NULL;
 	self->prev = NULL;
 #endif
+
 	g_datalist_clear (&self->object_data);
+	g_datalist_clear (&self->object_data_nocp);
 	
 	if (self->redraw_pending) {
 		self->free_pending = TRUE;
@@ -92,12 +95,15 @@ copy (HTMLObject *self,
 	dest->percent = self->percent;
 	dest->flags = self->flags;
 	dest->redraw_pending = FALSE;
-	dest->selected = self->selected;
+	dest->selected = FALSE;
 	dest->free_pending = FALSE;
 	dest->change = self->change;
+	dest->draw_focused = FALSE;
 
 	g_datalist_init (&dest->object_data);
 	html_object_copy_data_from_object (dest, self);
+
+	g_datalist_init (&dest->object_data_nocp);
 }
 
 static HTMLObject *
@@ -128,7 +134,10 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 		} else
 			html_object_move_cursor_before_remove (self, e);
 
-		html_object_change_set   (self,  HTML_CHANGE_ALL_CALC);
+		html_object_change_set (self, HTML_CHANGE_ALL_CALC);
+		html_object_change_set (self->parent, HTML_CHANGE_ALL_CALC);
+		/* force parent redraw */
+		self->parent->width = 0;
 		html_object_remove_child (self->parent, self);
 		*len += html_object_get_recursive_length (self);
 
@@ -140,6 +149,11 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 static gboolean
 merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **right, HTMLCursor *cursor)
 {
+	if (self->parent) {
+		html_object_change_set (self->parent, HTML_CHANGE_ALL_CALC);
+		self->parent->width = 0;
+	}
+
 	return FALSE;
 }
 
@@ -191,27 +205,6 @@ draw (HTMLObject *o,
 	/* Do nothing by default.  We don't know how to paint ourselves.  */
 }
 
-static void
-draw_background (HTMLObject *self,
-		 HTMLPainter *p,
-		 gint x, gint y,
-		 gint width, gint height,
-		 gint tx, gint ty)
-{
-	/* By default, objects are transparent so they simply forward
-           this to the parent.  */
-	if (self->parent != NULL) {
-		html_object_draw_background (self->parent, p,
-					     x + self->parent->x,
-					     y + self->parent->y - self->parent->ascent,
-					     width, height,
-					     tx - self->parent->x,
-					     ty - self->parent->y + self->parent->ascent);
-	} else {
-		/* FIXME this should draw the default background somehow.  */
-	}
-}
-
 static gboolean
 is_transparent (HTMLObject *self)
 {
@@ -226,12 +219,11 @@ fit_line (HTMLObject *o,
 	  gboolean next_to_floating,
 	  gint width_left)
 {
-	html_object_calc_size (o, painter, FALSE);
 	return (o->width <= width_left || (first_run && !next_to_floating)) ? HTML_FIT_COMPLETE : HTML_FIT_NONE;
 }
 
 static gboolean
-calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
+html_object_real_calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
 	return FALSE;
 }
@@ -287,13 +279,13 @@ reset (HTMLObject *o)
 }
 
 static const gchar *
-get_url (HTMLObject *o)
+get_url (HTMLObject *o, gint offset)
 {
 	return NULL;
 }
 
 static const gchar *
-get_target (HTMLObject *o)
+get_target (HTMLObject *o, gint offset)
 {
 	return NULL;
 }
@@ -322,10 +314,13 @@ static GdkColor *
 get_bg_color (HTMLObject *o,
 	      HTMLPainter *p)
 {
-	
-	return o->parent
-		? html_object_get_bg_color (o->parent, p)
-		: &((html_colorset_get_color (p->color_set, HTMLBgColor))->color);
+	if (o->parent)
+		return html_object_get_bg_color (o->parent, p);	
+
+	if (p->widget && GTK_IS_HTML (p->widget))
+		return &((html_colorset_get_color (GTK_HTML (p->widget)->engine->settings->color_set,
+						   HTMLBgColor))->color);
+	return NULL;
 }
 
 static HTMLObject*
@@ -498,7 +493,7 @@ select_range (HTMLObject *self,
 	gboolean selected;
 	gboolean changed;
 
-	selected = length > 0 || (length == -1 && start < html_object_get_length (self)) ? TRUE : FALSE;
+	selected = length > 0 || (length == -1 && start < html_object_get_length (self)) || html_object_is_container (self) ? TRUE : FALSE;
 	changed  = (! selected && self->selected) || (selected && ! self->selected) ? TRUE : FALSE;
 
 	self->selected = selected;
@@ -550,8 +545,7 @@ save_plain (HTMLObject *self,
 }
 
 static gint
-check_page_split (HTMLObject *self,
-		  gint y)
+check_page_split (HTMLObject *self, HTMLPainter *p, gint y)
 {
 	return 0;
 }
@@ -621,10 +615,9 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->remove_child = remove_child;
 	klass->split = split;
 	klass->draw = draw;
-	klass->draw_background = draw_background;
 	klass->is_transparent = is_transparent;
 	klass->fit_line = fit_line;
-	klass->calc_size = calc_size;
+	klass->calc_size = html_object_real_calc_size;
 	klass->set_max_width = set_max_width;
 	klass->set_max_height = set_max_height;
 	klass->get_left_margin = get_left_margin;
@@ -696,8 +689,10 @@ html_object_init (HTMLObject *o,
 	o->redraw_pending = FALSE;
 	o->free_pending = FALSE;
 	o->selected = FALSE;
+	o->draw_focused = FALSE;
 
 	g_datalist_init (&o->object_data);
+	g_datalist_init (&o->object_data_nocp);
 }
 
 HTMLObject *
@@ -867,16 +862,6 @@ html_object_draw (HTMLObject *o,
 	(* HO_CLASS (o)->draw) (o, p, x, y, width, height, tx, ty);
 }
 
-void
-html_object_draw_background (HTMLObject *o,
-			     HTMLPainter *p,
-			     gint x, gint y,
-			     gint width, gint height,
-			     gint tx, gint ty)
-{
-	(* HO_CLASS (o)->draw_background) (o, p, x, y, width, height, tx, ty);
-}
-
 gboolean
 html_object_is_transparent (HTMLObject *self)
 {
@@ -980,24 +965,24 @@ html_object_get_uris (HTMLObject *o, char **link, char **target, char **src)
 #endif 
 
 const gchar *
-html_object_get_url (HTMLObject *o)
+html_object_get_url (HTMLObject *o, gint offset)
 {
-	return (* HO_CLASS (o)->get_url) (o);
+	return (* HO_CLASS (o)->get_url) (o, offset);
 }
 
 const gchar *
-html_object_get_target (HTMLObject *o)
+html_object_get_target (HTMLObject *o, gint offset)
 {
-	return (* HO_CLASS (o)->get_target) (o);
+	return (* HO_CLASS (o)->get_target) (o, offset);
 }
 
 gchar *
-html_object_get_complete_url (HTMLObject *o)
+html_object_get_complete_url (HTMLObject *o, gint offset)
 {
 	const gchar *url, *target;
 
-	url = html_object_get_url (o);
-	target = html_object_get_target (o);
+	url = html_object_get_url (o, offset);
+	target = html_object_get_target (o, offset);
 	return url || target ? g_strconcat (url ? url : "#", url ? (target && *target ? "#" : NULL) : target,
 					      url ? target : NULL, NULL) : NULL;
 }
@@ -1236,12 +1221,11 @@ html_object_save_plain (HTMLObject *self,
 }
 
 gint
-html_object_check_page_split  (HTMLObject *self,
-			       gint y)
+html_object_check_page_split  (HTMLObject *self, HTMLPainter *p, gint y)
 {
 	g_return_val_if_fail (self != NULL, 0);
 
-	return (* HO_CLASS (self)->check_page_split) (self, y);
+	return (* HO_CLASS (self)->check_page_split) (self, p, y);
 }
 
 void
@@ -1604,6 +1588,24 @@ html_object_get_index (HTMLObject *self, guint offset)
 	return html_object_is_text (self) ? html_text_get_index (HTML_TEXT (self), offset) : offset;
 }
 
+void
+html_object_set_data_nocp (HTMLObject *object, const gchar *key, const gchar *value)
+{
+	g_datalist_set_data_full (&object->object_data_nocp, key, g_strdup (value), g_free);
+}
+
+void
+html_object_set_data_full_nocp (HTMLObject *object, const gchar *key, const gpointer value, GDestroyNotify func)
+{
+	g_datalist_set_data_full (&object->object_data_nocp, key, value, func);
+}
+
+gpointer
+html_object_get_data_nocp (HTMLObject *object, const gchar *key)
+{
+	return g_datalist_get_data (&object->object_data_nocp, key);
+}
+
 void
 html_object_set_data (HTMLObject *object, const gchar *key, const gchar *value)
 {
@@ -1648,8 +1650,8 @@ object_save_data (GQuark key_id, gpointer data, gpointer user_data)
 	/* printf ("object %s %s\n", g_quark_to_string (key_id), str); */
 	if (!str) {
 		/* printf ("save %s %s -> %s\n", state->save_data_class_name, g_quark_to_string (key_id), (gchar *) data); */
-		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" key=\"%s\" value=\"%s\">",
-						state->save_data_class_name, g_quark_to_string (key_id), data);
+		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" key=\"%s\" value=\"%s\">-->",
+						state->save_data_class_name, g_quark_to_string (key_id), (char *) data);
 		html_engine_set_class_data (state->engine, state->save_data_class_name, g_quark_to_string (key_id), data);
 	}
 }
@@ -1664,13 +1666,13 @@ handle_object_data (gpointer key, gpointer value, gpointer data)
 	/* printf ("handle: %s %s %s %s\n", state->save_data_class_name, key, value, str); */
 	if (!str) {
 		/* printf ("clear\n"); */
-		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" clear=\"%s\">",
-						state->save_data_class_name, key);
+		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" clear=\"%s\">-->",
+						state->save_data_class_name, (char *) key);
 		state->data_to_remove = g_slist_prepend (state->data_to_remove, key);
 	} else if (strcmp (value, str)) {
 		/* printf ("change\n"); */
-		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" key=\"%s\" value=\"%s\">",
-						state->save_data_class_name, key, str);
+		html_engine_save_output_string (state, "<!--+GtkHTML:<DATA class=\"%s\" key=\"%s\" value=\"%s\">-->",
+						state->save_data_class_name, (char *) key, str);
 		html_engine_set_class_data (state->engine, state->save_data_class_name, key, value);
 	}
 }
@@ -1822,19 +1824,6 @@ html_object_get_head_leaf (HTMLObject *o)
 	return rv;
 }
 
-static void
-clear_word_width (HTMLObject *o, HTMLEngine *e, gpointer data)
-{
-	if (html_object_is_text (o))
-		html_text_clear_word_width (HTML_TEXT (o));
-}
-
-void
-html_object_clear_word_width (HTMLObject *o)
-{
-	html_object_forall (o, NULL, clear_word_width, NULL);
-}
-
 HTMLObject *
 html_object_nth_parent (HTMLObject *self, gint n)
 {
@@ -2023,7 +2012,7 @@ next_prev_cursor_object (HTMLObject *o, HTMLEngine *e, gint *offset, gboolean fo
 	HTMLCursor cursor;
 	gboolean result;
 
-	html_cursor_init (&cursor, o, HTML_IS_TABLE (o) ? *offset : (forward ? html_object_get_length (o) : 0));
+	html_cursor_init (&cursor, o, html_object_is_container (o) ? *offset : (forward ? html_object_get_length (o) : 0));
 
 	result = forward ? html_cursor_forward (&cursor, e) : html_cursor_backward (&cursor, e);
 	*offset = cursor.offset;
